from typing import Dict, List, Callable, Any
from ml_gym.loss_functions.loss_functions import Loss, LossWarmupMixin
from ml_gym.models.nn.net import NNModel
from ml_gym.data_handling.dataset_loader import DatasetLoader
import torch
from ml_gym.batching.batch import InferenceResultBatch, DatasetBatch, Batch
from ml_gym.gym.inference_component import InferenceComponent
from ml_gym.gym.stateful_components import StatefulComponent
from torch.optim.optimizer import Optimizer
import tqdm
from ml_gym.util.logger import LogLevel, ConsoleLogger


class TrainComponent(StatefulComponent):
    def __init__(self, inference_component: InferenceComponent, loss_fun: Loss, show_progress: bool = False):
        self.loss_fun = loss_fun
        self.inference_component = inference_component
        self.show_progress = show_progress
        self.logger = ConsoleLogger("logger_train_component")

    def train_batch(self, batch: DatasetBatch, model: NNModel, optimizer: Optimizer, device: torch.device):
        model.zero_grad()
        batch.to_device(device)
        loss = self.calc_loss(model, batch)
        loss.sum().backward()
        optimizer.step()

    def train_epoch(self, model: NNModel, optimizer: Optimizer, data_loader: DatasetLoader,
                    device: torch.device) -> NNModel:
        self.map_batches(fun=self.train_batch,
                         loader=data_loader,
                         fun_params={"device": device,
                                     "model": model,
                                     "optimizer": optimizer},
                         show_progress=self.show_progress)
        return model

    def warm_up(self, model: NNModel, data_loader: DatasetLoader, device: torch.device):
        def init_loss(loss_fun: Loss, batch: InferenceResultBatch):
            if isinstance(loss_fun, LossWarmupMixin):
                loss_fun.warm_up(batch)
                loss_fun.finish_warmup()

        def check_if_initiable_component(loss_fun: Loss):
            return isinstance(loss_fun, LossWarmupMixin)

        if check_if_initiable_component(self.loss_fun):
            self.logger.log(LogLevel.INFO, "Running warmup...")
            with torch.no_grad():
                prediction_batches = self.map_batches(fun=self.forward_batch,
                                                      fun_params={"device": device, "model": model},
                                                      loader=data_loader,
                                                      show_progress=self.show_progress)
            prediction_batch = InferenceResultBatch.combine(prediction_batches)
            init_loss(self.loss_fun, prediction_batch)
        else:
            self.logger.log(LogLevel.INFO, "Skipping training warmup. No special loss functions to be initialized.")

    def forward_batch(self, batch: DatasetBatch, model: NNModel, device: torch.device) -> InferenceResultBatch:
        model.to(device)
        batch.to_device(device)
        forward_result_batch = self.inference_component.predict(model, batch)
        return forward_result_batch

    def calc_loss(self, model: NNModel, batch: DatasetBatch) -> torch.Tensor:
        forward_batch = self.inference_component.predict(model, batch)
        loss = self.loss_fun(forward_batch)
        return loss

    @staticmethod
    def map_batches(fun: Callable[[DatasetBatch, NNModel], Any], loader: DatasetLoader,
                    fun_params: Dict[str, Any] = None, show_progress: bool = False) -> List[Batch]:
        """
        Applies a function to each batch within a DatasetLoader
        """
        # TODO: This loads the entire dataset into memory.
        # We should make this a generator instead to prevent memory overflows.
        # Also code duplication with evaluator
        fun_params = fun_params if fun_params is not None else dict()
        if show_progress:
            return [fun(batch, **fun_params) for batch in tqdm.tqdm(loader, desc="Batches processed:")]
        else:
            return [fun(batch, **fun_params) for batch in loader]


class Trainer(StatefulComponent):
    def __init__(self, train_component: TrainComponent, train_loader: DatasetLoader, verbose=False):
        self.train_component = train_component
        self.train_loader = train_loader
        self.verbose = verbose

    def warm_up(self, model: NNModel, device: torch.device):
        self.train_component.warm_up(model, self.train_loader, device)

    def train_epoch(self, model: NNModel, optimizer: Optimizer, device: torch.device) -> NNModel:
        model = self.train_component.train_epoch(model, optimizer, self.train_loader, device)
        return model



from typing import List, Type, Dict, Any
from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.gym.gym import Gym
from ml_gym.validation.validator_factory import ValidatorFactory
from datetime import datetime
from ml_gym.util.logger import QueuedLogging
from multiprocessing import Queue
from shutil import copyfile
import os
from ml_gym.io.config_parser import YAMLConfigLoader
from ml_gym.error_handling.exception import ValidationModeNotValidError
from enum import Enum


class MLGymStarter:

    class ValidationMode(Enum):
        NESTED_CV = "nested_cv"
        GRID_SEARCH = "grid_search"

    def __init__(self, blue_print_class: Type[BluePrint], validation_mode: ValidationMode, num_epochs: int, dashify_logging_path: str,
                 gs_config_path: str, evaluation_config_path: str, text_logging_path: str, process_count: int,
                 gpus: List[int], log_std_to_file: bool) -> None:
        self.blue_print_class = blue_print_class
        self.num_epochs = num_epochs
        self.validation_mode = validation_mode
        self.dashify_logging_path = dashify_logging_path
        self.evaluation_config_path = evaluation_config_path
        self.text_logging_path = text_logging_path
        self.process_count = process_count
        self.log_std_to_file = log_std_to_file
        self.gpus = gpus
        self.gs_config_path = gs_config_path

    @staticmethod
    def _create_gym(process_count: int, device_ids, log_std_to_file: bool) -> Gym:
        gym = Gym(process_count, device_ids=device_ids, log_std_to_file=log_std_to_file)
        return gym

    @staticmethod
    def _setup_logging_environment(log_dir_path: str):
        queue = Queue()
        QueuedLogging.start_logging(queue, log_dir_path)

    @staticmethod
    def _stop_logging_environment():
        QueuedLogging.stop_listener()

    @staticmethod
    def _save_gs_config(gs_config_path: str, dashify_logging_path: str, grid_search_id: str):
        gs_logging_path = os.path.join(dashify_logging_path, grid_search_id)
        os.makedirs(gs_logging_path, exist_ok=True)
        copyfile(gs_config_path, os.path.join(gs_logging_path, os.path.basename(gs_config_path)))

    @staticmethod
    def _save_evaluation_config(evaluation_config_path: str, dashify_logging_path: str, grid_search_id: str):
        gs_logging_path = os.path.join(dashify_logging_path, grid_search_id)
        os.makedirs(gs_logging_path, exist_ok=True)
        copyfile(evaluation_config_path, os.path.join(gs_logging_path, os.path.basename(evaluation_config_path)))

    def start(self):
        grid_search_id = datetime.now().strftime("%Y-%m-%d--%H-%M-%S")
        self._save_gs_config(self.gs_config_path, self.dashify_logging_path, grid_search_id)
        self._setup_logging_environment(self.text_logging_path)
        gym = MLGymStarter._create_gym(process_count=self.process_count, device_ids=self.gpus, log_std_to_file=self.log_std_to_file)
        gs_config = YAMLConfigLoader.load(self.gs_config_path)
        if self.validation_mode == MLGymStarter.ValidationMode.NESTED_CV:
            self._save_evaluation_config(self.evaluation_config_path, self.dashify_logging_path, grid_search_id)
            evaluation_config = YAMLConfigLoader.load(self.evaluation_config_path)
            self.run_nested_cv(gym=gym, gs_config=gs_config, cv_config=evaluation_config, grid_search_id=grid_search_id)
        elif self.validation_mode == MLGymStarter.ValidationMode.GRID_SEARCH:
            self.run_grid_search(gym=gym, gs_config=gs_config, grid_search_id=grid_search_id)
        else:
            raise ValidationModeNotValidError
        self._stop_logging_environment()

    def run_nested_cv(self, gym: Gym, cv_config: Dict[str, Any], gs_config: Dict[str, Any], grid_search_id: str):
        nested_cv = ValidatorFactory.get_nested_cv(gs_config=gs_config,
                                                   cv_config=cv_config,
                                                   grid_search_id=grid_search_id,
                                                   blue_print_type=self.blue_print_class)
        nested_cv.run(blue_print_type=self.blue_print_class,
                      gym=gym,
                      gs_config=gs_config,
                      num_epochs=self.num_epochs,
                      dashify_logging_path=self.dashify_logging_path)

    def run_grid_search(self, gym: Gym, gs_config: Dict[str, Any], grid_search_id: str):
        gs_validator = ValidatorFactory.get_gs_validator(grid_search_id=grid_search_id)
        gs_validator.run(blue_print_type=self.blue_print_class,
                         gym=gym,
                         gs_config=gs_config,
                         num_epochs=self.num_epochs,
                         dashify_logging_path=self.dashify_logging_path)



import torch
from dashify.logging.dashify_logging import ExperimentTracking
from typing import List
from ml_gym.multiprocessing.pool import Pool, Job
from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.gym.jobs import AbstractGymJob
from functools import partial
from ml_gym.util.devices import get_devices
import tqdm

torch.multiprocessing.set_start_method('spawn', force=True)


class Gym:
    def __init__(self, process_count: int = 1, device_ids: List[int] = None, log_std_to_file: bool = True):
        self.devices = get_devices(device_ids)
        self.log_std_to_file = log_std_to_file
        self.pool = Pool(num_processes=process_count, devices=self.devices)
        self.jobs: List[Job] = []

    def run(self, parallel=True):
        """Executes the jobs. Note that this function blocks until all jobs have been executed!

        Args:
            parallel (bool, optional): When set to True, jobs are run in parallel in the multiprocessing environment. Defaults to True.
        """
        if parallel:
            for job in self.jobs:
                self.pool.add_job(job)
            self.pool.run()
        else:
            self.work(self.devices[0])

    def add_blue_print(self, blue_print: BluePrint):
        job = Job(job_id=len(self.jobs), fun=Gym._run_job, param_dict={"blue_print": blue_print, "log_std_to_file": self.log_std_to_file})
        self.jobs.append(job)

    def add_blue_prints(self, blue_prints: List[BluePrint]):
        for blue_print in blue_prints:
            self.add_blue_print(blue_print)

    @staticmethod
    def _run_job(blue_print: BluePrint, device: torch.device, log_std_to_file: bool) -> AbstractGymJob:
        gym_job = AbstractGymJob.from_blue_print(blue_print)
        decorated_runner = ExperimentTracking(gym_job.experiment_info, log_to_file=log_std_to_file)(partial(gym_job.execute, device=device))
        decorated_runner(device=device)
        return gym_job

    def work(self, device: torch.device):
        for job in tqdm.tqdm(self.jobs, desc="Models trained"):
            job.device = device
            job.execute()



from ml_gym.gym.jobs import AbstractGymJob
from ml_gym.blueprints.blue_prints import BluePrint
from typing import List, Type, Dict, Any
from ml_gym.util.grid_search import GridSearch


def create_blueprints(blue_print_class: Type[BluePrint],
                      run_mode: AbstractGymJob.Mode,
                      gs_config: Dict[str, Any],
                      dashify_logging_path: str,
                      num_epochs: int,
                      grid_search_id: str,
                      external_injection: Dict[str, Any] = None,
                      initial_experiment_id: int = 0) -> List[BluePrint]:

    run_id_to_config_dict = {run_id: config for run_id, config in enumerate(
        GridSearch.create_gs_from_config_dict(gs_config))}
    config_tuples = list(run_id_to_config_dict.items())
    epochs = list(range(num_epochs))

    blueprints = []
    for run_id, config in config_tuples:
        blue_print = blue_print_class(grid_search_id=grid_search_id,
                                      run_id=str(run_id + initial_experiment_id),
                                      epochs=epochs,
                                      run_mode=run_mode,
                                      config=config,
                                      dashify_logging_dir=dashify_logging_path,
                                      external_injection=external_injection)
        blueprints.append(blue_print)
    return blueprints



from typing import List
from ml_gym.gym.post_processing import PredictPostProcessingIF
from ml_gym.batching.batch import InferenceResultBatch


class PredictPostprocessingComponent:

    @staticmethod
    def post_process(result_batch: InferenceResultBatch, post_processors: List[PredictPostProcessingIF]) -> InferenceResultBatch:
        for post_processor in post_processors:
            result_batch = post_processor.postprocess(result_batch)
        return result_batch



from ml_gym.models.nn.net import NNModel
from ml_gym.batching.batch import InferenceResultBatch, DatasetBatch
from typing import List
import torch
from ml_gym.gym.predict_postprocessing_component import PredictPostprocessingComponent
from ml_gym.gym.post_processing import PredictPostProcessingIF
import tqdm
from ml_gym.data_handling.dataset_loader import DatasetLoader


class ExportedModel:
    def __init__(self, model: NNModel, post_processors: List[PredictPostProcessingIF]):
        self.post_processors = post_processors
        self.model = model

    def predict_tensor(self, sample_tensor: torch.Tensor):
        with torch.no_grad():
            forward_result = self.model.forward(sample_tensor)
        result_batch = InferenceResultBatch(targets=None, tags=None, predictions=forward_result)
        return PredictPostprocessingComponent.post_process(result_batch, post_processors=self.post_processors)

    def predict_dataset_batch(self, batch: DatasetBatch) -> InferenceResultBatch:
        with torch.no_grad():
            forward_result = self.model.forward(batch.samples)
        result_batch = InferenceResultBatch(targets=batch.targets, tags=batch.tags, predictions=forward_result)
        return PredictPostprocessingComponent.post_process(result_batch, post_processors=self.post_processors)

    def predict_data_loader(self, dataset_loader: DatasetLoader) -> InferenceResultBatch:
        result_batches = [self.predict_dataset_batch(batch) for batch in tqdm.tqdm(dataset_loader, desc="Batches processed:")]
        return InferenceResultBatch.combine(result_batches)


class InferenceComponent:
    def __init__(self, post_processors: List[PredictPostProcessingIF], no_grad=True):
        self.post_processors = post_processors
        self.no_grad = no_grad

    def predict(self, model: NNModel, batch: DatasetBatch) -> InferenceResultBatch:
        if self.no_grad:
            with torch.no_grad():
                forward_result = model.forward(batch.samples)
        else:
            forward_result = model.forward(batch.samples)
        result_batch = InferenceResultBatch(targets=batch.targets, tags=batch.tags, predictions=forward_result)
        return PredictPostprocessingComponent.post_process(result_batch, post_processors=self.post_processors)

    def predict_data_loader(self, model: NNModel, dataset_loader: DatasetLoader) -> InferenceResultBatch:
        result_batches = [self.predict(model, batch) for batch in tqdm.tqdm(dataset_loader, desc="Batches processed:")]
        return InferenceResultBatch.combine(result_batches)

    # TODO, this is really just a temporary hack and should be done at a different place!
    def export_model(self, model: NNModel) -> ExportedModel:
        return ExportedModel(model, self.post_processors)



from typing import Callable, Dict, Any, Union, List
import torch
from sklearn.metrics import roc_auc_score, average_precision_score
from ml_gym.batching.batch import InferenceResultBatch
from abc import ABC, abstractmethod
import numpy as np
from torch import nn


def binary_auroc_score(y_true: torch.Tensor, y_pred: torch.Tensor, **params: Dict[str, Any]) -> float:
    score = roc_auc_score(y_true=y_true, y_score=y_pred, **params)
    return score


def binary_aupr_score(y_true: torch.Tensor, y_pred: torch.Tensor, **params: Dict[str, Any]) -> float:
    score = average_precision_score(y_true=y_true, y_score=y_pred, **params)
    return score


class MetricIF(ABC):

    @abstractmethod
    def __call__(self, result_batch: InferenceResultBatch) -> Any:
        raise NotImplementedError


class Metric(MetricIF):

    def __init__(self, tag: str, identifier: str):
        self.tag = tag
        self.identifier = identifier

    @abstractmethod
    def __call__(self, result_batch: InferenceResultBatch) -> float:
        raise NotImplementedError


class PredictionMetric(Metric):

    def __init__(self, tag: str, identifier: str, target_subscription_key: str,
                 prediction_subscription_key: str, metric_fun: Callable, params: Dict[str, Any]):
        super().__init__(tag=tag, identifier=identifier)
        self.target_subscription_key = target_subscription_key
        self.prediction_subscription_key = prediction_subscription_key
        self.metric_fun = metric_fun
        self.params = params

    def __call__(self, result_batch: InferenceResultBatch) -> float:
        y_true = result_batch.get_targets(self.target_subscription_key).cpu()
        y_pred = result_batch.get_predictions(self.prediction_subscription_key).cpu()
        return self.metric_fun(y_true=y_true, y_pred=y_pred, **self.params)


class ClassSpecificExpectedCalibrationErrorMetric(Metric):

    def __init__(self, tag: str, identifier: str, target_subscription_key: str,
                 prediction_subscription_key: str, num_bins: int = 10, class_label: int = 1, sum_up_bins: bool = True):
        super().__init__(tag=tag, identifier=identifier)
        self.target_subscription_key = target_subscription_key
        self.prediction_subscription_key = prediction_subscription_key
        self.num_bins = num_bins
        self.class_label = class_label
        self.bins = np.linspace(0, 1, self.num_bins+1)[1:-1]
        self.sum_up_bins = sum_up_bins

    def _calc_prevalence(self, y_true: torch.Tensor) -> torch.Tensor:
        # all positive samples / total population (within a single bin)
        return torch.sum(y_true == self.class_label)/float(len(y_true))

    def _calc_confidence(self, y_pred: torch.Tensor) -> torch.Tensor:
        # average confidence over all samples
        return torch.sum(y_pred) / len(y_pred)

    def _calc_calibration_error(self, y_true: torch.Tensor, y_pred: torch.Tensor) -> float:
        ece_score = torch.abs(self._calc_prevalence(y_true) - self._calc_confidence(y_pred))
        return float(ece_score)

    def __call__(self, result_batch: InferenceResultBatch) -> Union[float, List[float]]:
        y_true = result_batch.get_targets(self.target_subscription_key).cpu().flatten()
        y_pred = result_batch.get_predictions(self.prediction_subscription_key).cpu().flatten()  # confidence for single class

        # bin samples based on prediction confidence
        bin_indices = np.digitize(y_pred.detach().numpy(), bins=self.bins)
        ce_scores = []
        bin_weights = []
        for bin_indice in range(self.num_bins):
            bin_mask = bin_indices == bin_indice
            y_pred_bin = y_pred[bin_mask]
            y_true_bin = y_true[bin_mask]
            if len(y_pred_bin) > 0:
                ce_bin_score = self._calc_calibration_error(y_true=y_true_bin, y_pred=y_pred_bin)
                bin_weight = len(y_pred_bin)/len(y_pred)
                ce_scores.append(ce_bin_score)
                bin_weights.append(bin_weight)
            else:
                ce_scores.append(0)
                bin_weights.append(0)
        if self.sum_up_bins:
            ece_score = sum([score * weight for score, weight in zip(ce_scores, bin_weights)])
            return ece_score
        else:
            return ce_scores


class BrierScoreMetric(Metric):
    def __init__(self, tag: str, identifier: str,
                 prediction_subscription_key: str,
                 target_subscription_key: str):
        super().__init__(tag=tag, identifier=identifier)
        self.target_subscription_key = target_subscription_key
        self.prediction_subscription_key = prediction_subscription_key
        self.mse = nn.MSELoss(reduction="mean")

    def __call__(self, inference_result_batch: InferenceResultBatch) -> float:
        y_true = inference_result_batch.get_targets(self.target_subscription_key).cpu()
        y_pred = inference_result_batch.get_predictions(self.prediction_subscription_key).cpu()
        return self.mse(y_pred.flatten(), y_true.flatten()).item()



import torch
import torch.nn as nn
from abc import abstractmethod
from typing import Dict


class NNModel(nn.Module):
    def __init__(self, seed: int = None):
        if seed is not None:
            torch.manual_seed(seed)
        super(NNModel, self).__init__()

    @abstractmethod
    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError

    @abstractmethod
    def forward_impl(self, inputs: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError

    def get_parameters(self) -> Dict[str, torch.Tensor]:
        return {name: param for name, param in self.named_parameters()}



from abc import ABC, abstractmethod
import torch
import torch.nn as nn
from ml_gym.loss_functions.loss_scaler import MeanScaler
from ml_gym.batching.batch import InferenceResultBatch
from typing import List, Callable
from ml_gym.gym.stateful_components import StatefulComponent
import torch.nn.functional as F
from ml_gym.error_handling.exception import InvalidTensorFormatError


class LossWarmupMixin(StatefulComponent):
    def __init__(self):
        self.warmup_loss_tensor: torch.Tensor = torch.Tensor()

    @abstractmethod
    def warm_up(self, forward_batch: InferenceResultBatch) -> torch.Tensor:
        raise NotImplementedError

    @abstractmethod
    def finish_warmup(self):
        raise NotImplementedError


class Loss(ABC):
    def __init__(self, tag: str = ""):
        self._tag = tag

    @property
    def tag(self) -> str:
        return self._tag

    @abstractmethod
    def __call__(self, eval_batch: InferenceResultBatch) -> torch.Tensor:
        """
        Calculates the loss
        :return: Loss tensor
        """
        raise NotImplementedError


class LPLoss(Loss):
    def __init__(self, target_subscription_key: str, prediction_subscription_key: str, root: int = 1, exponent: int = 2,
                 sample_selection_fun: Callable[[InferenceResultBatch], List[bool]] = None,
                 tag: str = "", average_batch_loss: bool = False):
        super().__init__(tag)
        self.root = root
        self.exponent = exponent
        self.target_subscription_key = target_subscription_key
        self.prediction_subscription_key = prediction_subscription_key
        self.sample_selection_fun = sample_selection_fun
        self.average_batch_loss = average_batch_loss

    def __call__(self, forward_batch: InferenceResultBatch) -> torch.Tensor:
        # here: predictions are reconstructions and targets are the input vectors
        t = forward_batch.get_targets(self.target_subscription_key)
        p = forward_batch.get_predictions(self.prediction_subscription_key)
        if t.shape != p.shape:
            raise InvalidTensorFormatError

        if self.sample_selection_fun is not None:
            sample_selection_mask = self.sample_selection_fun(forward_batch)
            t = t[sample_selection_mask]
            p = p[sample_selection_mask]
        loss_values = (torch.sum((p - t).abs() ** self.exponent, dim=1) ** (1 / self.root))
        if self.average_batch_loss:
            loss_values = torch.sum(loss_values)/len(loss_values)
        return loss_values


class LPLossScaled(LPLoss, LossWarmupMixin):
    def __init__(self, target_subscription_key: str, prediction_subscription_key: str, root: int = 1, exponent: int = 2,
                 sample_selection_fun: Callable[[InferenceResultBatch], List[bool]] = None, tag: str = ""):
        LPLoss.__init__(self, target_subscription_key,
                        prediction_subscription_key, root, exponent, sample_selection_fun, tag)
        LossWarmupMixin.__init__(self)
        self.warmup_losses = []
        self.scaler = MeanScaler()

    def __call__(self, forward_batch: InferenceResultBatch) -> torch.Tensor:
        loss_tensor = LPLoss.__call__(self, forward_batch)
        loss_tensor_scaled = self.scaler.scale(loss_tensor)
        return loss_tensor_scaled

    def warm_up(self, forward_batch: InferenceResultBatch) -> torch.Tensor:
        # calculate losses
        loss_tensor = self(forward_batch)
        self.warmup_losses.append(loss_tensor)
        return loss_tensor

    def finish_warmup(self):
        loss_tensor = torch.cat(self.warmup_losses)
        self.scaler.train(loss_tensor)


class CrossEntropyLoss(Loss):

    def __init__(self, target_subscription_key: str, prediction_subscription_key: str, tag: str = ""):
        super().__init__(tag)
        self.target_subscription_key: str = target_subscription_key
        self.prediction_subscription_key: str = prediction_subscription_key

    def __call__(self, inference_result_batch: InferenceResultBatch) -> torch.Tensor:
        # the targets tensor has each target in a separate tensor torch.Tensor([[1], [2], ..., [1]).
        # For the CrossEntropyLoss API we need them to be squeezed torch.Tensor([1, 2, ..., 1])
        t = inference_result_batch.get_targets(self.target_subscription_key).long()
        p = inference_result_batch.get_predictions(self.prediction_subscription_key)
        loss_values = nn.CrossEntropyLoss(reduction="none")(p, t)
        return loss_values


class NLLLoss(Loss):
    """The negative log likelihood loss.
    NOTE: Obtaining log-probabilities in a neural network is easily achieved by adding a LogSoftmax layer 
    in the last layer of your network. Here, we already did this implicitly in the loss function. 
    Therefore, this loss function is equivalent to CrossEntropyLoss"""

    def __init__(self, target_subscription_key: str, prediction_subscription_key: str, tag: str = ""):
        super().__init__(tag)
        self.target_subscription_key: str = target_subscription_key
        self.prediction_subscription_key: str = prediction_subscription_key

    def __call__(self, inference_result_batch: InferenceResultBatch) -> torch.Tensor:
        t = inference_result_batch.get_targets(self.target_subscription_key).long()
        p = inference_result_batch.get_predictions(self.prediction_subscription_key)
        p = F.log_softmax(p, dim=1)
        loss_values = nn.NLLLoss(reduction="none")(p, t)
        return loss_values


class BCEWithLogitsLoss(Loss):
    def __init__(self, target_subscription_key: str, prediction_subscription_key: str, tag: str = "", average_batch_loss: bool = False):
        super().__init__(tag)
        self.target_subscription_key: str = target_subscription_key
        self.prediction_subscription_key: str = prediction_subscription_key
        self.average_batch_loss = average_batch_loss

    def __call__(self, inference_result_batch: InferenceResultBatch) -> torch.Tensor:
        t = inference_result_batch.get_targets(
            self.target_subscription_key).float().flatten()
        p = inference_result_batch.get_predictions(
            self.prediction_subscription_key).flatten()
        loss_values = nn.BCEWithLogitsLoss(reduction="none")(p, t)
        if self.average_batch_loss:
            loss_values = torch.sum(loss_values)/len(loss_values)
        return loss_values



from abc import ABC, abstractmethod
from ml_gym.gym.evaluator import EvaluationBatchResult
from typing import List, Dict, Any
from dashify.logging.dashify_logging import ExperimentInfo, DashifyLogger


class AbstractWriter(ABC):
    @staticmethod
    @abstractmethod
    def log_console(result: EvaluationBatchResult):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def log_measurement_result(results: List[EvaluationBatchResult], experiment_info: ExperimentInfo,
                               measurement_id: int):
        raise NotImplementedError


class DashifyWriter(AbstractWriter):
    """ Partially wraps the `DashifyLogger`.
    """

    @staticmethod
    def log_console(result: EvaluationBatchResult):
        result.aggregate()
        print(str(result))

    @staticmethod
    def save_state(data_dict: Dict[str, Any], experiment_info: ExperimentInfo, measurement_id: int = 0):
        file_name = f"state_{str(measurement_id)}.json"
        DashifyLogger.save_dict(file_name, data_dict, experiment_info)

    @staticmethod
    def log_measurement_result(results: List[EvaluationBatchResult], experiment_info: ExperimentInfo,
                               measurement_id: int):
        results_dict = {}
        for result in results:
            metrics_dictionary = DashifyWriter._get_results_dict(result)
            results_dict = {**metrics_dictionary, **results_dict}
        DashifyLogger.log_metrics(experiment_info=experiment_info, metrics=results_dict, measurement_id=measurement_id)

    @staticmethod
    def save_model_state(model_state: Dict, experiment_info: ExperimentInfo, measurement_id: int):
        DashifyLogger.save_checkpoint_state_dict(model_state, "model", experiment_info, measurement_id)

    @staticmethod
    def save_optimizer_state(optimizer_state: Dict, experiment_info: ExperimentInfo, measurement_id: int):
        DashifyLogger.save_checkpoint_state_dict(optimizer_state, "optimizer", experiment_info, measurement_id)

    @staticmethod
    def _get_results_dict(result: EvaluationBatchResult) -> Dict[str, List[float]]:
        def combine_metric_name(dataset_split: str, metric_name: str):
            return f"{dataset_split}/{metric_name.lower()}"

        # get metrics and losses as dict
        results_dict = dict()
        metrics_and_losses = {**result.metrics, **result.losses}
        for metric_name, metric_value in metrics_and_losses.items():
            results_dict[combine_metric_name(result.split_name, metric_name)] = metric_value
        return results_dict


class DashifyReader:
    @staticmethod
    def load_model_state(experiment_info: ExperimentInfo, measurement_id: int) -> Dict:
        model_state = DashifyLogger.load_checkpoint_state_dict("model", experiment_info, measurement_id)
        return model_state

    @staticmethod
    def load_optimizer_state(experiment_info: ExperimentInfo, measurement_id: int) -> Dict:
        optimizer_state = DashifyLogger.load_checkpoint_state_dict("optimizer", experiment_info, measurement_id)
        return optimizer_state

    @staticmethod
    def load_state(experiment_info: ExperimentInfo, measurement_id: int = 0) -> Dict[str, Any]:
        file_name = f"state_{str(measurement_id)}.json"
        data_dict = DashifyLogger.load_dict(file_name, experiment_info)
        return data_dict



from typing import Dict, Any, Tuple, List, Type
from data_stack.dataset.iterator import DatasetIteratorIF
from ml_gym import create_blueprints
from ml_gym.gym.gym import Gym
from ml_gym.gym.jobs import AbstractGymJob
from data_stack.dataset.splitter import SplitterFactory
from ml_gym.validation.validator import ValidatorIF
from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.blueprints.component_factory import Injector


class NestedCV(ValidatorIF):
    def __init__(self, dataset_iterator: DatasetIteratorIF, num_outer_loop_folds: int,
                 num_inner_loop_folds: int, inner_stratification: bool, outer_stratification: bool,
                 target_pos: int, shuffle: bool, grid_search_id: str, seed: int):
        self.num_outer_loop_folds = num_outer_loop_folds
        self.num_inner_loop_folds = num_inner_loop_folds
        self.inner_stratification = inner_stratification
        self.outer_stratification = outer_stratification
        self.dataset_iterator = dataset_iterator
        self.seed = seed
        self.grid_search_id = grid_search_id
        self.target_pos = target_pos
        self.shuffle = shuffle

    def _get_fold_indices(self) -> Tuple[List[int]]:
        splitter = SplitterFactory.get_nested_cv_splitter(num_inner_loop_folds=self.num_inner_loop_folds,
                                                          num_outer_loop_folds=self.num_outer_loop_folds,
                                                          inner_stratification=self.inner_stratification,
                                                          outer_stratification=self.outer_stratification,
                                                          target_pos=self.target_pos,
                                                          shuffle=self.shuffle,
                                                          seed=self.seed)
        indices = splitter.get_indices(dataset_iterator=self.dataset_iterator)
        return indices

    def _create_blue_prints(self, blue_print_type: Type[BluePrint], gs_config: Dict[str, Any], num_epochs: int, dashify_logging_path: str):
        outer_fold_indices, inner_folds_indices = self._get_fold_indices()
        blueprints = []
        # outer folds
        # [outer_fold_1_indices, outer_fold2_indices ...]
        for outer_fold_id, test_fold_indices in enumerate(outer_fold_indices):
            train_fold_indices = []
            for i, fold in enumerate(outer_fold_indices):
                if i != outer_fold_id:
                    train_fold_indices = train_fold_indices + fold

            external_injection = {
                "id_split_indices": {
                    "train": train_fold_indices,
                    "test": test_fold_indices
                },
                "id_fold_tags": {"outer_test_fold_id": outer_fold_id,
                                 "inner_test_fold_id": -1}
            }
            injector = Injector(mapping=external_injection)
            gs_config_injected = injector.inject_pass(component_parameters=gs_config)
            bp = create_blueprints(blue_print_class=blue_print_type,
                                   run_mode=AbstractGymJob.Mode.TRAIN,
                                   gs_config=gs_config_injected,
                                   dashify_logging_path=dashify_logging_path,
                                   num_epochs=num_epochs,
                                   # external_injection=external_injection,
                                   grid_search_id=self.grid_search_id,
                                   initial_experiment_id=len(blueprints))
            blueprints = blueprints + bp

        # inner folds
        # [[outer_fold_1_inner_fold_1_indices, outer_fold_1_inner_fold_2_indices, ...], ...]
        for outer_fold_id, folds in enumerate(inner_folds_indices):
            for inner_fold_id, test_fold_indices in enumerate(folds):
                # calc train fold indices
                train_fold_indices = []
                for i, fold in enumerate(folds):
                    if i != inner_fold_id:
                        train_fold_indices = train_fold_indices + fold

                external_injection = {
                    "id_split_indices": {
                        "train": train_fold_indices,
                        "test": test_fold_indices
                    },
                    "id_fold_tags": {"outer_test_fold_id": outer_fold_id,
                                     "inner_test_fold_id": inner_fold_id}
                }
                injector = Injector(mapping=external_injection)
                gs_config_injected = injector.inject_pass(component_parameters=gs_config)
                bp = create_blueprints(blue_print_class=blue_print_type,
                                       run_mode=AbstractGymJob.Mode.TRAIN,
                                       gs_config=gs_config_injected,
                                       dashify_logging_path=dashify_logging_path,
                                       num_epochs=num_epochs,
                                       # external_injection=external_injection,
                                       grid_search_id=self.grid_search_id,
                                       initial_experiment_id=len(blueprints))
                blueprints = blueprints + bp
        return blueprints

    def run(self, blue_print_type: Type[BluePrint], gym: Gym, gs_config: Dict[str, Any], num_epochs: int, dashify_logging_path: str):
        blueprints = self._create_blue_prints(blue_print_type=blue_print_type,
                                              gs_config=gs_config,
                                              dashify_logging_path=dashify_logging_path,
                                              num_epochs=num_epochs)
        gym.add_blue_prints(blueprints)
        gym.run(parallel=True)
